/*
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#define SND_MAIN_OBJECT_FILE
#include "driver.h"

#if 0
#define DEBUG_ACCESS( d, m ) printk( ">>" d "[%i]\n", m )
#else
#define DEBUG_ACCESS( d, m )
#endif

int snd_major = SNDCFG_MAJOR;
int snd_config_dma_static = 0;
#ifdef MODULE_PARM              /* hey - we have new 2.1.18+ kernel... */
MODULE_AUTHOR( "Jaroslav Kysela <perex@jcu.cz>" );
MODULE_DESCRIPTION( "Ultra Sound Low-level driver for soundcards." );
MODULE_SUPPORTED_DEVICE( "sound" );
MODULE_PARM( snd_major, "i" );
MODULE_PARM_DESC( snd_major, "Major # for Ultra driver. (14 by default)" );
MODULE_PARM( snd_config_dma_static, "i" );
MODULE_PARM_DESC( snd_config_dma_static, "Enable static DMA memory allocation. (off by default)" );
#endif

#define VERIFY_CARD( idx, minor ) \
  ( snd_verify_map[ idx ] && !(snd_cards_bitmap & (1 << ((minor >> 4) & 0x07))) )

#define SND_MINOR_OSS		0x10

static char *snd_comment_map[ 32 ];
static int snd_verify_map[ 32 ];
static snd_unregister_t *snd_unregister_map[ 32 ];
static snd_lseek_t *snd_lseek_map[ 32 ];
static snd_read_t *snd_read_map[ 32 ];
static snd_write_t *snd_write_map[ 32 ];
static snd_open_t *snd_open_map[ 32 ];
static snd_release_t *snd_release_map[ 32 ];
#ifdef SND_POLL
static snd_poll_t *snd_poll_map[ 32 ];
#else
static snd_select_t *snd_select_map[ 32 ];
#endif
static snd_ioctl_t *snd_ioctl_map[ 32 ];
static snd_mmap_t *snd_mmap_map[ 32 ];

#ifdef LINUX_2_1_X
static loff_t snd_lseek( struct file *file, loff_t offset, int orig )
#else
#  ifdef LINUX_2_1
   static long long snd_lseek( struct inode *inode, struct file *file, long long offset, int orig )
#  else
   static int snd_lseek( struct inode *inode, struct file *file, off_t offset, int orig )
#  endif
#endif /* LINUX_2_1_X */
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  snd_lseek_t *ptr;

  DEBUG_ACCESS( "lseek", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_lseek_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, offset, orig );
}    

#ifdef LINUX_2_1_X
static ssize_t snd_read( struct file *file, char *buf, size_t count, loff_t *offset )
#else
#  ifdef LINUX_2_1
   static long snd_read( struct inode *inode, struct file *file, char *buf, unsigned long count )
#  else
   static int snd_read( struct inode *inode, struct file *file, char *buf, int count )
#  endif
#endif
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  snd_read_t *ptr;

  DEBUG_ACCESS( "read", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_read_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, buf, count );
}

#ifdef LINUX_2_1_X
static ssize_t snd_write( struct file *file, const char *buf, size_t count, loff_t *offset )
#else
#  ifdef LINUX_2_1
   static long snd_write( struct inode *inode, struct file *file, const char *buf, unsigned long count )
#  else
   static int snd_write( struct inode *inode, struct file *file, const char *buf, int count )
#  endif
#endif
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( inode -> i_rdev );
#endif
  unsigned int idx;
  snd_write_t *ptr;

  DEBUG_ACCESS( "write", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_write_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, buf, count );
}

static int snd_open( struct inode *inode, struct file *file )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  snd_open_t *ptr;

  DEBUG_ACCESS( "open", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_open_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( minor, file );
}

static int snd_release( struct inode * inode, struct file *file )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  snd_release_t *ptr;

  DEBUG_ACCESS( "release", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_release_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( minor, file );
}

#ifdef SND_POLL
static unsigned int snd_poll( struct file *file, poll_table *wait )
{
#ifdef LINUX_2_1_X
  unsigned short minor = MINOR( file -> f_dentry -> d_inode -> i_rdev );
#else
  unsigned short minor = MINOR( file -> f_inode -> i_rdev );
#endif
  unsigned int idx;
  snd_poll_t *ptr;

  DEBUG_ACCESS( "poll", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_poll_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, wait );
}
#else
static int snd_select( struct inode *inode, struct file *file, 
                       int sel_type, select_table *wait )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  snd_select_t *ptr;

  DEBUG_ACCESS( "select", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_select_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, sel_type, wait );
}                       
#endif

static int snd_ioctl( struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg )
{
  unsigned short minor = MINOR( inode -> i_rdev );
  unsigned int idx;
  snd_ioctl_t *ptr;

  DEBUG_ACCESS( "ioctl", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_ioctl_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( file, cmd, arg );
}

#ifdef LINUX_2_1_X
static int snd_mmap( struct file *file, struct vm_area_struct *vma )
#else
static int snd_mmap( struct inode *inode, struct file *file, struct vm_area_struct *vma )
#endif
{
#ifdef LINUX_2_1_X
  struct inode *inode = file -> f_dentry -> d_inode;
  unsigned short minor = MINOR( inode -> i_rdev );
#else
  unsigned short minor = MINOR( file -> f_inode -> i_rdev );
#endif
  unsigned int idx;
  snd_mmap_t *ptr;

  DEBUG_ACCESS( "mmap", minor );
  if ( minor >= SND_MINOR_UBEGIN ) {
    idx = minor & SND_MINOR_UDEVMASK;
  } else {
    idx = SND_MINOR_OSS | ( minor & SND_MINOR_OSS_MASK );
  }
  if ( VERIFY_CARD( idx, minor ) ) return -ENODEV;
  ptr = snd_mmap_map[ idx ];
  if ( !ptr ) return -ENXIO;
  return ptr( inode, file, vma );
}

static struct file_operations snd_fops = {
  snd_lseek,	/* (l)lseek */
  snd_read,	/* read */
  snd_write,	/* write */
  NULL,		/* readdir */
#ifdef SND_POLL
  snd_poll,	/* poll */
#else
  snd_select,	/* select */
#endif
  snd_ioctl,	/* ioctl */
  snd_mmap,	/* mmap */
  snd_open,	/* open */
#ifdef LINUX_2_1
  snd_release,	/* release */
#else
  (void (*)( struct inode * inode, struct file *file ))snd_release,
#endif
#ifdef LINUX_2_1_X
  NULL,		/* fsync */
  NULL,		/* fasync */
  NULL,		/* check_media_change */
  NULL,		/* revalidate */
  NULL,		/* lock */
#endif
};

#ifndef LINUX_2_1

int snd_ioctl_in( long *addr )
{
  if ( verify_area( VERIFY_READ, addr, sizeof( int ) ) != 0 )
    {
      printk( "snd_ioctl_in: bad address 0x%lx\n", (long)addr );
      return 0;
    }
  return get_fs_int( addr );
}

int snd_ioctl_out( long *addr, int value )
{
  int err;
  
  if ( ( err = verify_area( VERIFY_WRITE, addr, sizeof( int ) ) ) != 0 )
    return err;
  put_fs_int( value, addr );
  return 0;
}

#else

int snd_ioctl_in( long *addr )
{
  int value;
  
  get_user( value, addr );
  return value;
}

int snd_ioctl_out( long *addr, int value )
{
  int err;

  if ( ( err = verify_area( VERIFY_WRITE, addr, sizeof( int ) ) ) != 0 )
    return err;
  put_user( value, addr );
  return 0;
}

#endif /* !LINUX_2_1 */

static int snd_fill_minor_map( int idx, struct snd_stru_register_minor *reg )
{
  if ( reg ) {
    if ( snd_unregister_map[ idx ] )	/* Oops.. Already registered!!! */
      return -EBUSY;
  } else {
    if ( !snd_unregister_map[ idx ] ) /* Oops.. Already unregistered!!! */
      return 0;				/* Don't fail... */
  }

  if ( reg ) {
    snd_comment_map[ idx ] = snd_malloc_strdup( reg -> comment, SND_SP_KERNEL );
  } else {
    snd_free_str( snd_comment_map[ idx ] );
    snd_comment_map[ idx ] = NULL;
  }
  snd_verify_map[ idx ] = reg ? reg -> verify_card : 0;
  snd_unregister_map[ idx ] = reg ? reg -> snd_unregister : NULL;
  snd_lseek_map[ idx ] = reg ? reg -> snd_lseek : NULL;
  snd_read_map[ idx ] = reg ? reg -> snd_read : NULL;
  snd_write_map[ idx ] = reg ? reg -> snd_write : NULL;
  snd_open_map[ idx ] = reg ? reg -> snd_open : NULL;
  snd_release_map[ idx ] = reg ? reg -> snd_release : NULL;
#ifdef SND_POLL
  snd_poll_map[ idx ] = reg ? reg -> snd_poll : NULL;
#else
  snd_select_map[ idx ] = reg ? reg -> snd_select : NULL;
#endif
  snd_ioctl_map[ idx ] = reg ? reg -> snd_ioctl : NULL;
  snd_mmap_map[ idx ] = reg ? reg -> snd_mmap : NULL;
  return 0;
}

int snd_register_minor( struct snd_stru_register_minor *reg )
{
  int idx, idx1, minor, err;
  
  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> snd_minors[ idx ];
    if ( minor < 0 ) break;
    if ( ( err = snd_fill_minor_map( minor, reg ) ) < 0 ) {
      /* Oops.. Try deallocate all minors.. */
      for ( idx1 = 0; idx1 < idx; idx1++ ) {
        minor = reg -> snd_minors[ idx1 ];
        snd_fill_minor_map( minor, NULL );
      }
      return err;
    }
  }

  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> oss_minors[ idx ];
    if ( minor < 0 ) break;
    if ( ( err = snd_fill_minor_map( SND_MINOR_OSS | minor, reg ) ) < 0 ) {
      /* Oops.. Try deallocate all minors.. */
      for ( idx1 = 0; idx1 < idx; idx1++ ) {
        minor = reg -> oss_minors[ idx1 ];
        snd_fill_minor_map( SND_MINOR_OSS | minor, NULL );
      }
      for ( idx1 = 0; idx1 < 16; idx1++ ) {
        minor = reg -> snd_minors[ idx1 ];
        if ( minor < 0 ) break;
        snd_fill_minor_map( minor, NULL );
      }
      return err;
    }
  }

  return 0;
}

int snd_unregister_minor( struct snd_stru_register_minor *reg )
{
  int idx, minor;

  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> snd_minors[ idx ];
    if ( minor < 0 ) break;
    snd_fill_minor_map( minor, NULL );
  }
  for ( idx = 0; idx < 16; idx++ ) {
    minor = reg -> oss_minors[ idx ];
    if ( minor < 0 ) break;
    snd_fill_minor_map( SND_MINOR_OSS | minor, NULL );
  }
  return 0;
}

/*
 *  INFO PART
 */

void snd_minor_info( snd_info_buffer_t *buffer )
{
  unsigned long flags;
  int idx;

  snd_iprintf( buffer, "\nInstalled native devices:\n" );
  for ( idx = 0; idx < 16; idx++ ) {
    CLI( &flags );
    if ( !snd_comment_map[ idx ] ) continue;
    snd_iprintf( buffer, "  %02i [%s]: %s\n", idx, snd_verify_map[ idx ] ? "card" : "glob", snd_comment_map[ idx ] );
    STI( &flags );
  }
  snd_iprintf( buffer, "Installed OSS devices:\n" );
  for ( idx = 16; idx < 32; idx++ ) {
    CLI( &flags );
    if ( !snd_comment_map[ idx ] ) continue;
    snd_iprintf( buffer, "  %02i [%s]: %s\n", idx - 16, snd_verify_map[ idx ] ? "card" : "glob", snd_comment_map[ idx ] );
    STI( &flags );
  }
}

/*
 *  INIT PART
 */

int init_module( void )
{
  MEMSET( &snd_comment_map, 0, sizeof( snd_comment_map ) );
  MEMSET( &snd_verify_map, 0, sizeof( snd_verify_map ) );
  MEMSET( &snd_lseek_map, 0, sizeof( snd_lseek_map ) );
  MEMSET( &snd_read_map, 0, sizeof( snd_read_map ) );
  MEMSET( &snd_write_map, 0, sizeof( snd_write_map ) );
  MEMSET( &snd_open_map, 0, sizeof( snd_open_map ) );
  MEMSET( &snd_release_map, 0, sizeof( snd_release_map ) );
#ifdef SND_POLL
  MEMSET( &snd_poll_map, 0, sizeof( snd_poll_map ) );
#else
  MEMSET( &snd_select_map, 0, sizeof( snd_select_map ) );
#endif
  MEMSET( &snd_ioctl_map, 0, sizeof( snd_ioctl_map ) );
  MEMSET( &snd_mmap_map, 0, sizeof( snd_ioctl_map ) );

  if ( register_chrdev( snd_major, "snd", &snd_fops ) ) {
    PRINTK("snd: unable to register major device number %d\n", snd_major );
    return -EIO;
  }
  snd_malloc_init();
  snd_driver_init();
  if ( snd_info_register() < 0 ) {
    PRINTD( "snd: unable to register info minors\n" );
    /* Not fatal error */
  }
  if ( snd_control_register() < 0 ) {
    PRINTD( "snd: unable to register control minors\n" );
    /* Not fatal error */
  }
  return 0;
}

void cleanup_module( void )
{
  int idx;
  snd_unregister_t *unreg;
  
  for ( idx = 0; idx < 32; idx++ ) {
    unreg = snd_unregister_map[ idx ];
    if ( unreg )
      unreg();
  }
  for ( idx = 0; idx < 32; idx++ ) {
    unreg = snd_unregister_map[ idx ];
    if ( unreg )
      PRINTK( "snd: %s device minor %i not unregistered!!!\n", idx & 0x10 ? "OSS" : "Sound", idx & 0x0f );
  }
  snd_malloc_done();
  if ( unregister_chrdev( snd_major, "snd" ) != 0 )
    PRINTK( "snd: unable to unregister major device number %d\n", snd_major );
}
 